/*=============================================================================
    Spirit V1.5.0
    Copyright (c) 1998-2002 Spirit-Dev
    http://spirit.sourceforge.net/

    Permission to copy, use, modify, sell and distribute this software
    is granted provided this copyright notice appears in all copies.
    This software is provided "as is" without express or implied
    warranty, and with no claim as to its suitability for any purpose.

    See spirit.hpp for full copyright notices.
=============================================================================*/
#ifndef SPIRIT_CLOSURE_HPP
#define SPIRIT_CLOSURE_HPP

///////////////////////////////////////////////////////////////////////////////
#include "boost/spirit/core/parser.hpp"
#include "boost/spirit/core/composite.hpp"
#include "boost/spirit/attribute/parametric.hpp"

#include "boost/phoenix/closures.hpp"
#include "boost/phoenix/primitives.hpp"
#include "boost/phoenix/casts.hpp"
#include "boost/phoenix/operators.hpp"

#include <boost/static_assert.hpp>

///////////////////////////////////////////////////////////////////////////////
//
//  Spirit predefined maximum closure limit. This limit defines the maximum
//  number of elements a closure can hold. This number defaults to 3. The
//  actual maximum is rounded up in multiples of 3. Thus, if this value
//  is 4, the actual limit is 6. The ultimate maximum limit in this
//  implementation is 15.
//
//  It should NOT be greater than PHOENIX_LIMIT!
//
///////////////////////////////////////////////////////////////////////////////

#if !defined(SPIRIT_CLOSURE_LIMIT)
#define SPIRIT_CLOSURE_LIMIT PHOENIX_LIMIT
#endif

///////////////////////////////////////////////////////////////////////////////
//
// ensure SPIRIT_CLOSURE_LIMIT <= PHOENIX_LIMIT and SPIRIT_CLOSURE_LIMIT <= 15
//
///////////////////////////////////////////////////////////////////////////////
BOOST_STATIC_ASSERT(SPIRIT_CLOSURE_LIMIT <= PHOENIX_LIMIT);
BOOST_STATIC_ASSERT(SPIRIT_CLOSURE_LIMIT <= 15);

///////////////////////////////////////////////////////////////////////////////
namespace spirit
{
///////////////////////////////////////////////////////////////////////////////
//
//  closure_context class
//
///////////////////////////////////////////////////////////////////////////////
template <typename ClosureT>
class closure_context
{
public:

    typedef typename phoenix::tuple_element<0,
        typename ClosureT::tuple_t>::type attr_t;
    typedef ClosureT base_t;

    closure_context(ClosureT const& clos)
    : frame(clos) {}

    ~closure_context() {}

    template <typename ParserT, typename ScannerT>
    void pre_parse(ParserT const& /*p*/, ScannerT const& /*scan*/)
    {}

    template <typename ResultT, typename ParserT, typename ScannerT>
    ResultT& post_parse(ResultT& hit, ParserT const& /*p*/, ScannerT const& /*scan*/)
    { hit.value() = frame[phoenix::tuple_index<0>()]; return hit; }

private:

    phoenix::closure_frame<typename ClosureT::phoenix_closure_t> frame;
};

///////////////////////////////////////////////////////////////////////////////
//
//  closure_parser class
//
///////////////////////////////////////////////////////////////////////////////
template <typename ParserT, typename ClosureT>
struct closure_parser
:   public unary<ParserT, parser<closure_parser<ParserT, ClosureT> > >
{
    typedef closure_parser<ParserT, ClosureT>   self_t;
    typedef unary<ParserT, parser<self_t> >     base_t;
    typedef typename phoenix::tuple_element
        <0, typename ClosureT::tuple_t>::type   attr_t;

    template <typename ScannerT>
    struct result
    {
        typedef typename match_result<ScannerT, attr_t>::type type;
    };

    closure_parser(ParserT const& p, ClosureT const& clos_)
    : base_t(p), clos(clos_) {}

    template <typename ScannerT>
    typename parser_result<self_t, ScannerT>::type
    parse(ScannerT const& scan) const
    {
        typedef typename parser_result<self_t, ScannerT>::type result_t;
        phoenix::closure_frame<ClosureT> frame(clos);
        return this->subject().parse(scan);
    }

    ClosureT const& clos;
};

///////////////////////////////////////////////////////////////////////////////
//
//  closure class
//
///////////////////////////////////////////////////////////////////////////////
template <
        typename DerivedT
    ,   typename T0 = phoenix::nil_t
    ,   typename T1 = phoenix::nil_t
    ,   typename T2 = phoenix::nil_t

#if SPIRIT_CLOSURE_LIMIT > 3
    ,   typename T3 = phoenix::nil_t
    ,   typename T4 = phoenix::nil_t
    ,   typename T5 = phoenix::nil_t

#if SPIRIT_CLOSURE_LIMIT > 6
    ,   typename T6 = phoenix::nil_t
    ,   typename T7 = phoenix::nil_t
    ,   typename T8 = phoenix::nil_t

#if SPIRIT_CLOSURE_LIMIT > 9
    ,   typename T9 = phoenix::nil_t
    ,   typename T10 = phoenix::nil_t
    ,   typename T11 = phoenix::nil_t

#if SPIRIT_CLOSURE_LIMIT > 12
    ,   typename T12 = phoenix::nil_t
    ,   typename T13 = phoenix::nil_t
    ,   typename T14 = phoenix::nil_t

#endif
#endif
#endif
#endif
>
struct closure :
    public phoenix::closure<
        T0, T1, T2
#if SPIRIT_CLOSURE_LIMIT > 3
    ,   T3, T4, T5
#if SPIRIT_CLOSURE_LIMIT > 6
    ,   T6, T7, T8
#if SPIRIT_CLOSURE_LIMIT > 9
    ,   T9, T10, T11
#if SPIRIT_CLOSURE_LIMIT > 12
    ,   T12, T13, T14
#endif
#endif
#endif
#endif
    >
{
    typedef phoenix::closure<
            T0, T1, T2
#if SPIRIT_CLOSURE_LIMIT > 3
        ,   T3, T4, T5
#if SPIRIT_CLOSURE_LIMIT > 6
        ,   T6, T7, T8
#if SPIRIT_CLOSURE_LIMIT > 9
        ,   T9, T10, T11
#if SPIRIT_CLOSURE_LIMIT > 12
        ,   T12, T13, T14
#endif
#endif
#endif
#endif
        > phoenix_closure_t;

    typedef closure_context<DerivedT> context_t;

    ~closure() {}
};

///////////////////////////////////////////////////////////////////////////////
//
//  free_closure class
//
///////////////////////////////////////////////////////////////////////////////
template <
        typename DerivedT
    ,   typename T0 = phoenix::nil_t
    ,   typename T1 = phoenix::nil_t
    ,   typename T2 = phoenix::nil_t

#if SPIRIT_CLOSURE_LIMIT > 3
    ,   typename T3 = phoenix::nil_t
    ,   typename T4 = phoenix::nil_t
    ,   typename T5 = phoenix::nil_t

#if SPIRIT_CLOSURE_LIMIT > 6
    ,   typename T6 = phoenix::nil_t
    ,   typename T7 = phoenix::nil_t
    ,   typename T8 = phoenix::nil_t

#if SPIRIT_CLOSURE_LIMIT > 9
    ,   typename T9 = phoenix::nil_t
    ,   typename T10 = phoenix::nil_t
    ,   typename T11 = phoenix::nil_t

#if SPIRIT_CLOSURE_LIMIT > 12
    ,   typename T12 = phoenix::nil_t
    ,   typename T13 = phoenix::nil_t
    ,   typename T14 = phoenix::nil_t

#endif
#endif
#endif
#endif
>
struct free_closure :
    public phoenix::closure<
        T0, T1, T2
#if SPIRIT_CLOSURE_LIMIT > 3
    ,   T3, T4, T5
#if SPIRIT_CLOSURE_LIMIT > 6
    ,   T6, T7, T8
#if SPIRIT_CLOSURE_LIMIT > 9
    ,   T9, T10, T11
#if SPIRIT_CLOSURE_LIMIT > 12
    ,   T12, T13, T14
#endif
#endif
#endif
#endif
    >
{
    typedef phoenix::closure<
            T0, T1, T2
#if SPIRIT_CLOSURE_LIMIT > 3
        ,   T3, T4, T5
#if SPIRIT_CLOSURE_LIMIT > 6
        ,   T6, T7, T8
#if SPIRIT_CLOSURE_LIMIT > 9
        ,   T9, T10, T11
#if SPIRIT_CLOSURE_LIMIT > 12
        ,   T12, T13, T14
#endif
#endif
#endif
#endif
        > phoenix_closure_t;

    typedef closure_context<DerivedT> context_t;

    template <typename ParserT>
    closure_parser<ParserT, phoenix_closure_t>
    operator[](ParserT const& p) const
    {
        return closure_parser<ParserT, phoenix_closure_t>(p, *this);
    }
};

///////////////////////////////////////////////////////////////////////////////
//
//  overloads for chseq_p and str_p taking in phoenix actors
//
///////////////////////////////////////////////////////////////////////////////
template <typename ActorT>
struct container_begin
{
    typedef container_begin<ActorT> self_t;

    template <typename TupleT>
    struct result
    {
        typedef typename phoenix::actor_result<ActorT, TupleT>
            ::plain_type::iterator type;
    };

    container_begin(ActorT actor_)
    : actor(actor_) {}

    template <typename TupleT>
    typename phoenix::actor_result<self_t, TupleT>::type
    eval(TupleT const& /*args*/) const
    { return actor().begin(); }

    ActorT actor;
};

//////////////////////////////////
template <typename ActorT>
struct container_end
{
    typedef container_begin<ActorT> self_t;

    template <typename TupleT>
    struct result
    {
        typedef typename phoenix::actor_result<ActorT, TupleT>
            ::plain_type::iterator type;
    };

    container_end(ActorT actor_)
    : actor(actor_) {}

    template <typename TupleT>
    typename phoenix::actor_result<self_t, TupleT>::type
    eval(TupleT const& /*args*/) const
    { return actor().end(); }

    ActorT actor;
};

//////////////////////////////////
template <typename BaseT>
inline fchseq<
    phoenix::actor<container_begin<phoenix::actor<BaseT> > >,
    phoenix::actor<container_end<phoenix::actor<BaseT> > >
>
fchseq_p(phoenix::actor<BaseT> const& a)
{
    typedef phoenix::actor<container_begin<phoenix::actor<BaseT> > >
        container_begin_t;
    typedef phoenix::actor<container_end<phoenix::actor<BaseT> > >
        container_end_t;
    typedef fchseq<container_begin_t, container_end_t> result_t;

    return result_t(container_begin_t(a), container_end_t(a));
}

//////////////////////////////////
template <typename BaseT>
inline fstrlit<
    phoenix::actor<container_begin<phoenix::actor<BaseT> > >,
    phoenix::actor<container_end<phoenix::actor<BaseT> > >
>
fstr_p(phoenix::actor<BaseT> const& a)
{
    typedef phoenix::actor<container_begin<phoenix::actor<BaseT> > >
        container_begin_t;
    typedef phoenix::actor<container_end<phoenix::actor<BaseT> > >
        container_end_t;
    typedef fstrlit<container_begin_t, container_end_t> result_t;

    return result_t(container_begin_t(a), container_end_t(a));
}

///////////////////////////////////////////////////////////////////////////////
}   //  namespace Spirit

#endif
