/*=============================================================================
    Spirit V1.5.0
    Copyright (c) 1998-2002 Spirit-Dev
    http://spirit.sourceforge.net/

    Permission to copy, use, modify, sell and distribute this software
    is granted provided this copyright notice appears in all copies.
    This software is provided "as is" without express or implied
    warranty, and with no claim as to its suitability for any purpose.

    See spirit.hpp for full copyright notices.
=============================================================================*/
#ifndef SPIRIT_GROUPING_IPP
#define SPIRIT_GROUPING_IPP

#include "boost/spirit/utility/refactoring.hpp"

///////////////////////////////////////////////////////////////////////////////
namespace spirit
{

///////////////////////////////////////////////////////////////////////////////
namespace impl 
{

///////////////////////////////////////////////////////////////////////////////
//
//  tuple_result type computer 
//
//      This template class returns the type of a tuple containing all the 
//      return values of all parsers, which compose a given parser.
//
///////////////////////////////////////////////////////////////////////////////
template <typename CategoryT>
struct tuple_result {

    template <typename ParserT, typename ScannerT>
    struct result {
    
        typedef
            typename parser_result<ParserT, ScannerT>::type  
            result_t;
        typedef typename phoenix::make_tuple<result_t>::type type;
    };
};

template <>
struct tuple_result<binary_parser_category> {

    template <typename ParserT, typename ScannerT>
    struct result {
    
        typedef typename ParserT::left_t                left_t;
        typedef typename ParserT::right_t               right_t;
        typedef typename left_t::parser_category_t      left_category_t;
        typedef typename right_t::parser_category_t     right_category_t;
        
        typedef typename tuple_result<left_category_t>
            ::template result<left_t, ScannerT>::type   left_result_t;
        typedef typename tuple_result<right_category_t>
            ::template result<right_t, ScannerT>::type  right_result_t;
            
        typedef 
            typename phoenix::concat_tuples<left_result_t, right_result_t>::type
            type;
    };
};

///////////////////////////////////////////////////////////////////////////////
//
//  Specialization of the match result assignment functionality for tuples.
//  This allows for returning the first tuple member as the parser result of 
//  the grouping parser (see match.ipp)
//
///////////////////////////////////////////////////////////////////////////////
template <
    typename A, typename B, typename C
#if PHOENIX_LIMIT > 3
    , typename D, typename E, typename F
#if PHOENIX_LIMIT > 6
    , typename G, typename H, typename I 
#if PHOENIX_LIMIT > 9
    , typename J, typename K, typename L 
#if PHOENIX_LIMIT > 12
    , typename M, typename N, typename O
#endif
#endif
#endif
#endif
>
struct convert<phoenix::tuple<A, B, C
#if PHOENIX_LIMIT > 3
        , D, E, F
#if PHOENIX_LIMIT > 6
        , G, H, I
#if PHOENIX_LIMIT > 9
        , J, K, L
#if PHOENIX_LIMIT > 12
        , M, N, O
#endif
#endif
#endif
#endif
    > > 
{
    typedef phoenix::tuple<A, B, C
#if PHOENIX_LIMIT > 3
        , D, E, F
#if PHOENIX_LIMIT > 6
        , G, H, I
#if PHOENIX_LIMIT > 9
        , J, K, L
#if PHOENIX_LIMIT > 12
        , M, N, O
#endif
#endif
#endif
#endif
    > tuple_t;

    template <typename T>
    static T 
    to_result(tuple_t const &t) { return t[phoenix::tuple_index<0>()]; }
};

///////////////////////////////////////////////////////////////////////////////
//
//  embed_actor template
//
//      This is a helper template for avoiding copy construction of an actor to 
//      be attached to a parser.
//
///////////////////////////////////////////////////////////////////////////////
template <typename ActorT>
class embed_actor {

public:
    embed_actor(ActorT const &actor_)
    : actor(actor_) {}
    
    template <typename T>
    void operator() (T const &val) const
    {
        actor(val);
    }
    
    template <typename IteratorT>
    void operator() (IteratorT const &begin, IteratorT const &end) const
    {
        actor(begin, end);
    }

private:
    ActorT const &actor;
};

///////////////////////////////////////////////////////////////////////////////
//
//  grouped_parser parse helper template
//
//      The grouping should be done for binary_category type parsers only.
//      All other parsers are left without any change.
//
///////////////////////////////////////////////////////////////////////////////

template <
    typename ParserT, typename ScannerT, typename GroupingT, 
    typename CategoryT
>
struct grouped_parser 
:   public parser<grouped_parser<ParserT, ScannerT, GroupingT, CategoryT> >
{
    typedef CategoryT parser_category_t;
    
    static typename parser_result<GroupingT, ScannerT>::type 
    parse (ParserT const &p, ScannerT const &scan) 
    {
        return p.parse(scan);
    }
};

///////////////////////////////////////////////////////////////////////////////
//  
//  This template is the real workhorse behind the group_d generated parsers, 
//  if the parser to group is a binary parser.
//
//  The main work this template does, is to attach itself as an
//  semantic action recursivly to all sub-parsers of the original binary 
//  parser. This is done by the help of the nested_attach_action_d refactoring
//  generator. So every sub-parser calls the corresponding operator() functions
//  of this class, providing it with their parser result value(s). Because 
//  the parsing is done from left to right, these result values are copied
//  into the appropriate position of a result tuple.
//  After a successful match this tuple value is feeded into the usual
//  Spirit semantic action mechanics and is available from inside the semantic
//  action code.
//  
///////////////////////////////////////////////////////////////////////////////

template <typename ParserT, typename ScannerT, typename GroupingT>
struct grouped_parser<ParserT, ScannerT, GroupingT, binary_parser_category> 
:   public parser<
        grouped_parser<ParserT, ScannerT, GroupingT, binary_parser_category> >
{
    typedef typename parser_result<GroupingT, ScannerT>::type result_t;
    typedef typename result_t::attr_t tuple_t;
    typedef binary_parser_category parser_category_t;
    typedef 
        grouped_parser<ParserT, ScannerT, GroupingT, binary_parser_category>
        self_t;
        
    typedef typename parser_result<ParserT, ScannerT>::type embedded_result_t;
    
    // the resulting tuple should contain at least 2 elements
    BOOST_STATIC_ASSERT(tuple_t::length >= 2);

    result_t 
    parse (ParserT const &p, ScannerT const &scan) const
    {
        member = 0;             // start assigning to the first tuple member
        
        const attach_action_gen<self_nested_refactoring> nested_attach_action_d = 
            attach_action_gen<self_nested_refactoring>(self_nested_refactoring());
            
        typename ScannerT::iterator_t save = scan.first;
        embedded_result_t hit 
            =   nested_attach_action_d[
                    p[embed_actor<self_t>(*this)] // avoid copy construction
                ].parse(scan);

        if (hit) {
            SPIRIT_ASSERT(member == tuple_t::length);
            return scan.create_match(hit.length(), t, save, scan.first);
        }
        return scan.no_match();
    }
    
    template <typename T>
    void operator() (T const &val) const
    {
        SPIRIT_ASSERT(member < tuple_t::length);
        typedef 
            phoenix::impl::dyn_tuple_member<tuple_t::length, T> 
            access_tuple_t;
        *reinterpret_cast<T *>(access_tuple_t::access(t, member++)) = val;
    }
    
    template <typename IteratorT>
    void operator() (IteratorT const &begin, IteratorT const &end) const
    {
        SPIRIT_ASSERT(member < tuple_t::length - 1);
        typedef 
            phoenix::impl::dyn_tuple_member<tuple_t::length, IteratorT> 
            access_tuple_t;
        *reinterpret_cast<IteratorT *>(access_tuple_t::access(t, member++)) = begin;
        *reinterpret_cast<IteratorT *>(access_tuple_t::access(t, member++)) = end;
    }

private:
    mutable int member;
    mutable tuple_t t;
};

///////////////////////////////////////////////////////////////////////////////
}   // namespace impl

///////////////////////////////////////////////////////////////////////////////
}   //  namespace Spirit

#endif // SPIRIT_GROUPING_IPP
