/*=============================================================================
    Spirit V1.5.0
    Copyright (c) 1998-2002 Spirit-Dev
    http://spirit.sourceforge.net/

    Permission to copy, use, modify, sell and distribute this software
    is granted provided this copyright notice appears in all copies.
    This software is provided "as is" without express or implied
    warranty, and with no claim as to its suitability for any purpose.

    See spirit.hpp for full copyright notices.
=============================================================================*/
#if !defined(SPIRIT_PARSER_HPP)
#define SPIRIT_PARSER_HPP

///////////////////////////////////////////////////////////////////////////////
#include "boost/config.hpp"
#include "boost/spirit/core/scanner.hpp"

#if defined(BOOST_MSVC)
#include "boost/spirit/core/impl/msvc.hpp"
#endif

#if defined(BOOST_MSVC) && (BOOST_MSVC <= 1300)
#define SPIRIT_PARSER_RESULT_ARGS typename ScannerT, typename Pizza = nil_t
#else
#define SPIRIT_PARSER_RESULT_ARGS typename ScannerT
#endif

///////////////////////////////////////////////////////////////////////////////
namespace spirit
{
template <typename ParserT, typename ActionT>
class action;   //  forward declaration

///////////////////////////////////////////////////////////////////////////////
//
//	Parser categories
//
//      Helper template classes to distinguish different types of
//      parsers. The following categories are the most generic. More
//      specific types may inherit from these. Each parser has a typedef
//      parser_category_t that defines its category. By default, if one is
//      not specified, it will inherit from the base parser class which
//      typedefs its parser_category_t as plain_parser_category.
//
//          - plain parser has nothing special
//          - binary parser has subject a and b (e.g. alternative)
//          - unary parser has single subject  (e.g. kleene star)
//          - action parser has an attached action parser
//
///////////////////////////////////////////////////////////////////////////////
struct plain_parser_category {};
struct binary_parser_category       : plain_parser_category {};
struct unary_parser_category        : plain_parser_category {};
struct action_parser_category       : unary_parser_category {};

///////////////////////////////////////////////////////////////////////////////
//
//  parser_result class
//
///////////////////////////////////////////////////////////////////////////////
#if defined(BOOST_MSVC) && (BOOST_MSVC <= 1300)

SPIRIT_DEPENDENT_TEMPLATE_WRAPPER(parser_result_wrapper, result);

//////////////////////////////////
template <typename ParserT, typename ScannerT>
struct parser_result
{
    typedef typename impl::parser_result_wrapper<ParserT>
        ::template result_<ScannerT>::type type;
};

#else

//////////////////////////////////
template <typename ParserT, typename ScannerT>
struct parser_result
{
    typedef typename ParserT::template result<ScannerT>::type type;
};

#endif

///////////////////////////////////////////////////////////////////////////////
//
//  parser class
//
///////////////////////////////////////////////////////////////////////////////
template <typename DerivedT>
struct parser
{
    typedef DerivedT                embed_t;
    typedef DerivedT                derived_t;
    typedef plain_parser_category   parser_category_t;

    template <SPIRIT_PARSER_RESULT_ARGS>
    struct result
    {
        typedef typename match_result<ScannerT, nil_t>::type type;
    };

    DerivedT& derived()
    { return *static_cast<DerivedT*>(this); }

    DerivedT const& derived() const
    { return *static_cast<DerivedT const*>(this); }

    template <typename ActionT>
    action<DerivedT, ActionT>
    operator[](ActionT const& actor) const
    { return action<DerivedT, ActionT>(derived(), actor); }
};

///////////////////////////////////////////////////////////////////////////////
//
//  parse_info
//
//      Results returned by the free parse functions:
//
//      stop:   points to the final parse position (i.e parsing
//              processed the input up to this point).
//
//      hit:    true if parsing is successful. This may be full:
//              the parser consumed all the input, or partial:
//              the parser consumed only a portion of the input.
//
//      full:   true when we have a full hit (i.e the parser
//              consumed all the input.
//
//      length: The number of characters consumed by the parser.
//              This is valid only if we have a successful hit
//              (either partial or full). A negative value means
//              that the hit is unsucessful.
//
///////////////////////////////////////////////////////////////////////////////
template <typename IteratorT = char const*>
struct parse_info
{
    IteratorT   stop;
    bool        hit;
    bool        full;
    unsigned    length;

    parse_info(
        IteratorT const& stop_ = IteratorT(),
        bool hit_ = false,
        bool full_ = false,
        unsigned length_ = 0)
    : stop(stop_)
    , hit(hit_)
    , full(full_)
    , length(length_) {}

    template <typename ParseInfoT>
    parse_info(ParseInfoT const& pi)
    : stop(pi.stop)
    , hit(pi.hit)
    , full(pi.full)
    , length(pi.length) {}
};

///////////////////////////////////////////////////////////////////////////////
//
//  Generic parse function
//
///////////////////////////////////////////////////////////////////////////////
template <typename IteratorT, typename DerivedT>
parse_info<IteratorT>
parse(
    IteratorT const&        first,
    IteratorT const&        last,
    parser<DerivedT> const& p);

///////////////////////////////////////////////////////////////////////////////
//
//  Parse function for null terminated strings
//
///////////////////////////////////////////////////////////////////////////////
template <typename CharT, typename DerivedT>
parse_info<CharT const*>
parse(
    CharT const*            str,
    parser<DerivedT> const& p);

///////////////////////////////////////////////////////////////////////////////
}   //  namespace Spirit

#endif

#include "boost/spirit/core/impl/parser.ipp"
#undef SPIRIT_PARSER_RESULT_ARGS
