/*=============================================================================
    Spirit V1.5.0
    Copyright (c) 1998-2002 Spirit-Dev
    http://spirit.sourceforge.net/

    Permission to copy, use, modify, sell and distribute this software
    is granted provided this copyright notice appears in all copies.
    This software is provided "as is" without express or implied
    warranty, and with no claim as to its suitability for any purpose.

    See spirit.hpp for full copyright notices.
=============================================================================*/
#ifndef SPIRIT_PRIMITIVES_HPP
#define SPIRIT_PRIMITIVES_HPP

///////////////////////////////////////////////////////////////////////////////
#include "boost/spirit/core/parser.hpp"
#include "boost/spirit/core/impl/directives.ipp"
#include "boost/spirit/core/impl/primitives.ipp"
#include "boost/spirit/debug.hpp"
#include <boost/ref.hpp>

///////////////////////////////////////////////////////////////////////////////
namespace spirit
{
///////////////////////////////////////////////////////////////////////////////
//
//  char_parser class
//
///////////////////////////////////////////////////////////////////////////////
template <typename DerivedT>
struct char_parser : public parser<DerivedT>
{
    typedef DerivedT self_t;
    template <typename ScannerT>
    struct result {

        typedef typename match_result
        <
            ScannerT,
            typename ScannerT::value_t
        >::type type;
    };

    template <typename ScannerT>
    typename parser_result<self_t, ScannerT>::type
    parse(ScannerT const& scan) const
    {
        typedef typename parser_result<self_t, ScannerT>::type result_t;
        typedef typename ScannerT::value_t      value_t;
        typedef typename ScannerT::iterator_t   iterator_t;

        if (!scan.at_end())
        {
            value_t ch = *scan;
            if (this->derived().test(ch))
            {
                iterator_t save(scan.first);
                ++scan;
                return scan.create_match(1, ch, save, scan.first);
            }
        }
        return scan.no_match();
    }
};

///////////////////////////////////////////////////////////////////////////////
//
//  chlit class
//
///////////////////////////////////////////////////////////////////////////////
template <typename CharT = char>
struct chlit : public char_parser<chlit<CharT> >
{
    chlit(CharT ch_)
    : ch(ch_) {}

    template <typename T>
    bool test(T ch_) const
    { return ch_ == ch; }

    CharT   ch;
};

//////////////////////////////////
template <typename CharT>
inline chlit<CharT>
ch_p(CharT ch)
{ return chlit<CharT>(ch); }

///////////////////////////////////////////////////////////////////////////////
//
//  range class
//
///////////////////////////////////////////////////////////////////////////////
template <typename CharT = char>
struct range : public char_parser<range<CharT> >
{
    range(CharT first_, CharT last_)
    : first(first_), last(last_)
    { SPIRIT_ASSERT(first <= last); }

    template <typename T>
    bool test(T ch) const
    { return (CharT(ch) >= first) && (CharT(ch) <= last); }

    CharT   first;
    CharT   last;
};

//////////////////////////////////
template <typename CharT>
inline range<CharT>
range_p(CharT first, CharT last)
{ return range<CharT>(first, last); }

///////////////////////////////////////////////////////////////////////////////
//
//  chseq class
//
///////////////////////////////////////////////////////////////////////////////
template <typename IteratorT = char const*>
class chseq : public parser<chseq<IteratorT> >
{
public:

    typedef chseq<IteratorT> self_t;

    chseq(IteratorT first_, IteratorT last_)
    : first(first_), last(last_) {}

    chseq(IteratorT first_)
    : first(first_), last(impl::get_last(first_)) {}

    template <typename ScannerT>
    typename parser_result<self_t, ScannerT>::type
    parse(ScannerT const& scan) const
    {
        typedef typename boost::unwrap_reference<IteratorT>::type striter_t;
        typedef typename parser_result<self_t, ScannerT>::type result_t;
        return impl::string_parser_parse<result_t>(
            striter_t(first),
            striter_t(last),
            scan);
    }

private:

    IteratorT first;
    IteratorT last;
};

//////////////////////////////////
template <typename CharT>
inline chseq<CharT const*>
chseq_p(CharT const* str)
{ return chseq<CharT const*>(str); }

//////////////////////////////////
template <typename IteratorT>
inline chseq<IteratorT>
chseq_p(IteratorT first, IteratorT last)
{ return chseq<IteratorT>(first, last); }

///////////////////////////////////////////////////////////////////////////////
//
//  strlit class
//
///////////////////////////////////////////////////////////////////////////////
template <typename IteratorT = char const*>
class strlit : public parser<strlit<IteratorT> >
{
public:

    typedef strlit<IteratorT> self_t;

    strlit(IteratorT first, IteratorT last)
    : seq(first, last) {}

    strlit(IteratorT first)
    : seq(first) {}

    template <typename ScannerT>
    typename parser_result<self_t, ScannerT>::type
    parse(ScannerT const& scan) const
    {
        typedef typename parser_result<self_t, ScannerT>::type result_t;
        return impl::contiguous_parser_parse<result_t>
            (seq, scan, scan);
    }

private:

    chseq<IteratorT> seq;
};

//////////////////////////////////
template <typename CharT>
inline strlit<CharT const*>
str_p(CharT const* str)
{ return strlit<CharT const*>(str); }

//////////////////////////////////
template <typename IteratorT>
inline strlit<IteratorT>
str_p(IteratorT first, IteratorT last)
{ return strlit<IteratorT>(first, last); }

///////////////////////////////////////////////////////////////////////////////
//
//  Predefined parser primitives
//
///////////////////////////////////////////////////////////////////////////////
struct nothing_ : public parser<nothing_>
{
    typedef nothing_ self_t;

    nothing_() {}

    template <typename ScannerT>
    typename parser_result<self_t, ScannerT>::type
    parse(ScannerT const& scan) const
    { return scan.no_match(); }
};

//////////////////////////////////
nothing_ const nothing_p = nothing_();

///////////////////////////////////////////////////////////////////////////////
struct epsilon_ : public parser<epsilon_>
{
    typedef epsilon_ self_t;

    epsilon_() {}

    template <typename ScannerT>
    typename parser_result<self_t, ScannerT>::type
    parse(ScannerT const& scan) const
    { return scan.empty_match(); }
};

//////////////////////////////////
epsilon_ const epsilon_p = epsilon_();
epsilon_ const eps_p = epsilon_();

///////////////////////////////////////////////////////////////////////////////
struct anychar_ : public char_parser<anychar_>
{
    typedef anychar_ self_t;

    anychar_() {}

    template <typename CharT>
    bool test(CharT) const
    { return true; }
};

//////////////////////////////////
anychar_ const anychar_p = anychar_();

///////////////////////////////////////////////////////////////////////////////
struct alnum_ : public char_parser<alnum_>
{
    typedef alnum_ self_t;

    alnum_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::isalnum_(ch); }
};

//////////////////////////////////
alnum_ const alnum_p = alnum_();

///////////////////////////////////////////////////////////////////////////////
struct alpha_ : public char_parser<alpha_>
{
    typedef alpha_ self_t;

    alpha_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::isalpha_(ch); }
};

//////////////////////////////////
alpha_ const alpha_p = alpha_();

///////////////////////////////////////////////////////////////////////////////
struct cntrl_ : public char_parser<cntrl_>
{
    typedef cntrl_ self_t;

    cntrl_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::iscntrl_(ch); }
};

//////////////////////////////////
cntrl_ const cntrl_p = cntrl_();

///////////////////////////////////////////////////////////////////////////////
struct digit_ : public char_parser<digit_>
{
    typedef digit_ self_t;

    digit_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::isdigit_(ch); }
};

//////////////////////////////////
digit_ const digit_p = digit_();

///////////////////////////////////////////////////////////////////////////////
struct graph_ : public char_parser<graph_>
{
    typedef graph_ self_t;

    graph_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::isgraph_(ch); }
};

//////////////////////////////////
graph_ const graph_p = graph_();

///////////////////////////////////////////////////////////////////////////////
struct lower_ : public char_parser<lower_>
{
    typedef lower_ self_t;

    lower_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::islower_(ch); }
};

//////////////////////////////////
lower_ const lower_p = lower_();

///////////////////////////////////////////////////////////////////////////////
struct print_ : public char_parser<print_>
{
    typedef print_ self_t;

    print_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::isprint_(ch); }
};

//////////////////////////////////
print_ const print_p = print_();

///////////////////////////////////////////////////////////////////////////////
struct punct_ : public char_parser<punct_>
{
    typedef punct_ self_t;

    punct_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::ispunct_(ch); }
};

//////////////////////////////////
punct_ const punct_p = punct_();

///////////////////////////////////////////////////////////////////////////////
struct blank_ : public char_parser<blank_>
{
    typedef blank_ self_t;

    blank_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return ch == ' ' || ch == '\t'; }
    // some systems have isblank(), but some don't, so just do it manually.
};

//////////////////////////////////
blank_ const blank_p = blank_();

///////////////////////////////////////////////////////////////////////////////
struct space_ : public char_parser<space_>
{
    typedef space_ self_t;

    space_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::isspace_(ch); }
};

//////////////////////////////////
space_ const space_p = space_();

///////////////////////////////////////////////////////////////////////////////
struct upper_ : public char_parser<upper_>
{
    typedef upper_ self_t;

    upper_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::isupper_(ch); }
};

//////////////////////////////////
upper_ const upper_p = upper_();

///////////////////////////////////////////////////////////////////////////////
struct xdigit_ : public char_parser<xdigit_>
{
    typedef xdigit_ self_t;

    xdigit_() {}

    template <typename CharT>
    bool test(CharT ch) const
    { return impl::isxdigit_(ch); }
};

//////////////////////////////////
xdigit_ const xdigit_p = xdigit_();

///////////////////////////////////////////////////////////////////////////////
struct eol_ : public parser<eol_>
{
    //////////////////////////////////
    //  eol_ parser contributed
    //  by Martin Wille
    //////////////////////////////////

    typedef eol_ self_t;

    eol_() {}

    template <typename ScannerT>
    typename parser_result<self_t, ScannerT>::type
    parse(ScannerT const& scan) const
    {
        typename ScannerT::iterator_t save = scan.first;
        int len = 0;

        if (!scan.at_end() && *scan == '\r')    // CR
        {
            ++scan;
            ++len;
        }

        if (!scan.at_end() && *scan == '\n')    // LF
        {
            ++scan;
            ++len;
        }

        if (len)
            return scan.create_match(len, nil_t(), save, scan.first);
        return scan.no_match();
    }
};

//////////////////////////////////
eol_ const eol_p = eol_();

///////////////////////////////////////////////////////////////////////////////
struct end_ : public parser<end_>
{
    //////////////////////////////////
    //  eol_ parser contributed
    //  by Markus Schpflin
    //////////////////////////////////

    typedef end_ self_t;

    end_() {}

    template <typename ScannerT>
    typename parser_result<self_t, ScannerT>::type
    parse(ScannerT const& scan) const
    {
        if (scan.at_end())
            return scan.empty_match();
        return scan.no_match();
    }
};

//////////////////////////////////
end_ const end_p = end_();

///////////////////////////////////////////////////////////////////////////////
}   //  namespace Spirit

#endif
